/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2008 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */
package org.ow2.petals.component.framework.listener;

import java.util.concurrent.atomic.AtomicInteger;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.apache.commons.pool.BasePoolableObjectFactory;
import org.apache.commons.pool.impl.GenericObjectPool;
import org.ow2.petals.component.framework.AbstractComponent;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.util.StringHelper;

/**
 * The factory used by the {@link GenericObjectPool} to create new
 * {@link AbstractJBIListener} instances.
 * 
 * @author chamerling - EBM WebSourcing
 * @since 3.0
 * 
 */
public class JBIListenerPoolableObjectFactory extends BasePoolableObjectFactory {

	private AbstractComponent component;

	private Logger logger;

	final AtomicInteger workerNumber = new AtomicInteger(1);

	/**
	 * Creates a new instance of {@link JBIListenerPoolableObjectFactory}
	 * 
	 * @param component
	 * @param runtimeConfiguration
	 * @param logger
	 */
	public JBIListenerPoolableObjectFactory(AbstractComponent component) {
		this.component = component;
		this.logger = component.getLogger();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.apache.commons.pool.BasePoolableObjectFactory#makeObject()
	 */
	@Override
	public Object makeObject() throws Exception {
		return createAndInitJBIListener();
	}

	/**
	 * Instantiate and initialize a JBI listener. This JBI Listener is the one
	 * which is defined by the component developer.
	 * 
	 * @return
	 * @throws PEtALSCDKException
	 */
	private AbstractJBIListener createAndInitJBIListener()
			throws PEtALSCDKException {
		AbstractJBIListener jbiListener = null;
		String jbiListenerClassName = component.getComponentConfiguration().getJbiListenerClassName();

		this.logger
				.log(Level.FINE, "Creating a new JBIListener '"
						+ jbiListenerClassName + "' #"
						+ workerNumber.getAndIncrement());

		if (!StringHelper.isNullOrEmpty(jbiListenerClassName)) {
			try {
				ClassLoader componentClassLoader = this.getClass()
						.getClassLoader();
				Class<?> clazz = componentClassLoader
						.loadClass(jbiListenerClassName);
				jbiListener = (AbstractJBIListener) clazz.newInstance();
			} catch (InstantiationException e) {
				throw new PEtALSCDKException("Can't instanciate JBIListener : "
						+ jbiListenerClassName, e);
			} catch (IllegalAccessException e) {
				throw new PEtALSCDKException("Can't instanciate JBIListener : "
						+ jbiListenerClassName, e);
			} catch (ClassNotFoundException e) {
				throw new PEtALSCDKException("Can't find JBIListener class : "
						+ jbiListenerClassName, e);
			}
			jbiListener.init(component);
			jbiListener.init();
		} else {
			throw new PEtALSCDKException(
					"JBIListener class name can not be empty");
		}
		return jbiListener;
	}
}
