/**
 * PETALS - PETALS Services Platform. Copyright (c) 2005 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.component.framework.util;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.ConnectException;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.net.URLConnection;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.List;
import java.util.StringTokenizer;

import javax.wsdl.Definition;
import javax.wsdl.WSDLException;
import javax.wsdl.xml.WSDLReader;
import javax.xml.namespace.QName;
import javax.xml.transform.TransformerException;

import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.DOMBuilder;
import org.jdom.input.SAXBuilder;
import org.jdom.output.DOMOutputter;
import org.ow2.easywsdl.wsdl.WSDLFactory;
import org.ow2.easywsdl.wsdl.api.Description;
import org.ow2.easywsdl.wsdl.api.Endpoint;
import org.ow2.easywsdl.wsdl.api.InterfaceType;
import org.ow2.easywsdl.wsdl.api.Service;
import org.ow2.easywsdl.wsdl.api.WSDLWriter;
import org.ow2.easywsdl.wsdl.api.abstractItf.AbsItfDescription.WSDLVersionConstants;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.w3c.dom.Document;
import org.w3c.dom.DocumentFragment;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import com.ibm.wsdl.xml.WSDLReaderImpl;

/**
 * Util class to manipulate WSDL Document or Definition.
 * 
 * @author wjoseph - EBMWebsourcing
 * @since 1.0
 */
public final class WSDLHelper {

    public static final String WSDL20_NS = "http://www.w3.org/ns/wsdl";

    public static final String WSDL11_NS = "http://schemas.xmlsoap.org/wsdl/";

    public static final String SCHEMA_NS = "http://www.w3.org/2001/XMLSchema";

    public static final org.jdom.Namespace XSI = org.jdom.Namespace.getNamespace("xsi",
            "http://http://www.w3.org/2001/XMLSchema-instance");

    public static final org.jdom.Namespace WSDL20 = org.jdom.Namespace.getNamespace(WSDL20_NS);

    public static org.jdom.Namespace WSDL11 = org.jdom.Namespace.getNamespace(WSDL11_NS);

    public static org.jdom.Namespace SCHEMA = org.jdom.Namespace.getNamespace(SCHEMA_NS);

    public static org.jdom.Namespace PETALS_WSDL_EXTENSIONS = org.jdom.Namespace.getNamespace(
            "petalsWsdlExtensions", "http://petals.ow2.org/wsdlExtensions");

    private WSDLFactory wsdlFactory;

    private WSDLWriter wsdlWriter;

    private static WSDLHelper wsdlHelper;

    /**
     * Creates a new instance of {@link WSDLHelper}
     * 
     * @throws org.ow2.easywsdl.wsdl.api.WSDLException
     * 
     */
    private WSDLHelper() throws org.ow2.easywsdl.wsdl.api.WSDLException {
        this.wsdlFactory = WSDLFactory.newInstance();
        this.wsdlWriter = this.wsdlFactory.newWSDLWriter();
    }

    /**
     * Lazy singletion.
     * 
     * @return
     * @throws org.ow2.easywsdl.wsdl.api.WSDLException
     */
    public synchronized static WSDLHelper getInstance()
            throws org.ow2.easywsdl.wsdl.api.WSDLException {
        if (wsdlHelper == null) {
            wsdlHelper = new WSDLHelper();
        }
        return wsdlHelper;
    }

    /**
     * Creates a {@link Definition} from a {@link Document}
     * 
     * @param doc
     * @return
     * @throws WSDLException
     * @deprecated Use EasyWSDL API instead
     */
    @Deprecated
    public static final Definition createDefinitionFromDocument(final Document doc)
            throws WSDLException {
        final WSDLReader reader = new WSDLReaderImpl();
        return reader.readWSDL(null, doc);
    }

    /**
     * Create a dom document from a File
     * 
     * @param wsdlFile
     *            file
     * @return dom document
     * @deprecated Use {@link WSDLHelper#createDocumentFromWSDL(File, boolean)}
     */
    @Deprecated
    public static final Document createDocumentFromWSDL(final File wsdlFile) {
        URL url = null;
        Document doc = null;
        try {
            url = wsdlFile.toURI().toURL();
            doc = createDocumentFromWSDL(url);
        } catch (final Exception e) {
            e.printStackTrace();
        }

        return doc;
    }

    /**
     * Create a dom document from a File
     * 
     * @param wsdlFile
     *            file
     * @param wsdlImportsDownload
     *            If <code>true</code>, the external imports declared in the
     *            service WSDL will be downloaded, so they will be replaced by
     *            their content.
     * @return dom document
     * @deprecated Use EasyWSDL API instead
     */
    @Deprecated
    public static final Document createDocumentFromWSDL(final File wsdlFile,
            final boolean wsdlImportsDownload) {
        URL url = null;
        Document doc = null;
        try {
            url = wsdlFile.toURI().toURL();
            doc = createDocumentFromWSDL(url, wsdlImportsDownload);
        } catch (final Exception e) {
            e.printStackTrace();
        }

        return doc;
    }

    /**
     * Get a {@link Document} from a wsdl located on the given {@link URL}.
     * 
     * @param url
     * @return null if there is an error getting the wsdl or if the parsing
     *         failed
     * @throws Exception
     *             the Document can not be created
     * @deprecated Use {@link WSDLHelper#createDocumentFromWSDL(URL, boolean)}
     */
    @Deprecated
    public static final Document createDocumentFromWSDL(final URL url) throws PEtALSCDKException {
        return WSDLHelper.createDocumentFromWSDL(url, false);
    }

    /**
     * Get a {@link Document} from a wsdl located on the given {@link URL}.
     * 
     * @param url
     * @param wsdlImportsDownload
     *            If <code>false</code>, the external imports declared in the
     *            service WSDL won't be downloaded, so they won't be replaced by
     *            their content.
     * @return null if there is an error getting the wsdl or if the parsing
     *         failed
     * @throws Exception
     *             the Document can not be created
     */
    @Deprecated
    public static final Document createDocumentFromWSDL(final URL url,
            final boolean wsdlImportsDownload) throws PEtALSCDKException {
        org.w3c.dom.Document doc = null;
        org.jdom.Document jdomDocument = null;
        URLConnection connection = null;
        InputStream is = null;

        try {
            connection = url.openConnection();
            is = connection.getInputStream();

            final SAXBuilder builder = new SAXBuilder();

            // Is it a local file ?
            String installroot = null;
            URI wsdlURI = url.toURI();
            if (wsdlURI.getScheme().equalsIgnoreCase("file")) {
                File wsdlFile = new File(wsdlURI);
                if (wsdlFile.exists()) {
                    File wsdlParent = new File(wsdlFile.getParent());
                    if (wsdlParent.isDirectory()) {
                        installroot = wsdlParent.getAbsolutePath();
                    }
                }
            }

            jdomDocument = builder.build(is);

            // add imported document
            jdomDocument = WSDLHelper.addImportedDocuments(jdomDocument, installroot,
                    wsdlImportsDownload);

            // Create DOM Document
            final DOMOutputter dom = new DOMOutputter();
            doc = dom.output(jdomDocument);

        } catch (final Exception e) {
            throw new PEtALSCDKException("Can not get WSDL document for URL " + url, e);
        } finally {
            try {
                if (is != null) {
                    is.close();
                }
            } catch (final IOException e) {
                // Nothing to do
            }
        }
        return doc;
    }

    /**
     * Remove the "importedDocuments" element from the given document
     * 
     * @param doc
     * @return
     * @throws PEtALSCDKException
     * @deprecated Use EasyWSDL API instead
     */
    @Deprecated
    public static final Document removeImportedDocuments(Document doc) throws PEtALSCDKException {
        Document newDoc = null;
        if (doc != null) {
            newDoc = (Document) doc.cloneNode(true);
            Node root = newDoc.getFirstChild();
            Node importedDocRoot = XMLUtil.findChild(root, "importedDocuments",
                    PETALS_WSDL_EXTENSIONS.getURI(), false);
            if (importedDocRoot != null) {
                root.removeChild(importedDocRoot);
            }
        }
        return newDoc;
    }

    /**
     * 
     * @param doc
     * @param path
     * @return
     * @throws PEtALSCDKException
     * @deprecated Use EasyWSDL API instead
     */
    @Deprecated
    public static final Document changeImportLocation(Document doc, final String path)
            throws PEtALSCDKException {
        try {
            if (doc != null) {
                // Create JDOM Document
                // Create new DOMBuilder, using default parser
                final DOMBuilder builder = new DOMBuilder();
                final org.jdom.Document jdomDoc = builder.build(doc);

                final org.jdom.Element root = jdomDoc.getRootElement();
                final String wsdlRoot = root.getName();
                if (((wsdlRoot.equals("definitions")) && (root.getNamespaceURI().equals(WSDL11_NS)))
                        || ((wsdlRoot.equals("description")) && (root.getNamespaceURI()
                                .equals(WSDL20_NS)))
                        || ((wsdlRoot.equals("schema")) && (root.getNamespaceURI()
                                .equals(SCHEMA_NS)))) {

                    final List<org.jdom.Element> imports = getAllImports(root);

                    // get in shemaLocation
                    if (root.getAttributeValue("schemaLocation", XSI) != null) {
                        String val = root.getAttributeValue("schemaLocation", XSI);
                        final StringTokenizer st = new StringTokenizer(val, " ");
                        String location = null;
                        String value = null;
                        while (st.hasMoreTokens()) {
                            st.nextToken();
                            value = null;
                            if (st.hasMoreTokens()) {
                                location = st.nextToken();
                                boolean isUrl = true;
                                try {
                                    new URL(location);
                                } catch (MalformedURLException e) {
                                    isUrl = false;
                                }
                                if (!isUrl) {
                                    value = path + location;
                                    val = val.replace(location, value);
                                }
                            }
                        }
                        root.getAttribute("schemaLocation", XSI).setValue(val);
                    }

                    if ((imports != null) && (imports.size() > 0)) {
                        for (final org.jdom.Element impt : imports) {
                            String value = impt.getAttributeValue("location");
                            if (value == null) {
                                value = impt.getAttributeValue("schemaLocation");
                            }

                            if (value != null) {
                                value = path + value;
                                if (impt.getAttribute("location") != null) {
                                    boolean isUrl = true;
                                    try {
                                        new URL(impt.getAttribute("location").getValue());
                                    } catch (MalformedURLException e) {
                                        isUrl = false;
                                    }
                                    if (!isUrl) {
                                        impt.getAttribute("location").setValue(value);
                                    }
                                } else if (impt.getAttribute("schemaLocation") != null) {
                                    boolean isUrl = true;
                                    try {
                                        new URL(impt.getAttribute("schemaLocation").getValue());
                                    } catch (MalformedURLException e) {
                                        isUrl = false;
                                    }
                                    if (!isUrl) {
                                        impt.getAttribute("schemaLocation").setValue(value);
                                    }
                                }
                            }
                        }

                        // Create DOM Document
                        final DOMOutputter dom = new DOMOutputter();
                        doc = dom.output(jdomDoc);
                    }
                }
            }
        } catch (final Exception e) {
            throw new PEtALSCDKException("Can not change import location in documents in WSDL", e);
        }
        return doc;
    }

    /**
     * Create a light WSDL2.0 document from the given parameters. All parameters
     * are required, otherwise a NullPointerException will be thrown. Namespaces
     * are managed.
     * 
     * @param interfaceName
     *            must be non null
     * @param serviceName
     *            must be non null
     * @param endpointName
     *            must be non null
     * @return a WSDL2.0 light document
     * @throws org.ow2.easywsdl.wsdl.api.WSDLException
     */
    public final Document createLightWSDL20(final QName interfaceName, final QName serviceName,
            final String endpointName)
            throws org.ow2.easywsdl.wsdl.api.WSDLException {

        Description wsdlDescription = wsdlFactory.newDescription(WSDLVersionConstants.WSDL20);

        //take the interface namespace as targetNamespace
        wsdlDescription.setTargetNamespace(interfaceName.getNamespaceURI());
        wsdlDescription.setQName(new QName("autoGenerated"));

        InterfaceType interfazz = wsdlDescription.createInterface();
        interfazz.setQName(interfaceName);

        Service service = wsdlDescription.createService();
        service.setQName(serviceName);
        service.setInterface(interfazz);

        Endpoint endpoint = service.createEndpoint();
        endpoint.setName(endpointName);
        service.addEndpoint(endpoint);
        
        wsdlDescription.addInterface(interfazz);
        wsdlDescription.addService(service);

        return wsdlWriter.getDocument(wsdlDescription);
    }

    /**
     * 
     * @param doc
     * @return
     * @throws PEtALSCDKException
     * @deprecated Use EasyWSDL API instead
     */
    @Deprecated
    public static final List<DocumentFragment> getImportedDocuments(Document doc) {
        final List<DocumentFragment> importedDocs = new ArrayList<DocumentFragment>();
        if (doc != null) {
            Node importedDocumentsNode = XMLUtil.findChild(doc, "importedDocuments",
                    PETALS_WSDL_EXTENSIONS.getURI(), true);
            if (importedDocumentsNode != null) {
                NodeList importedDocuments = importedDocumentsNode.getChildNodes();
                while (importedDocuments.getLength() > 0) {
                    Node child = importedDocuments.item(0);
                    // only add Element node (no Text elements)
                    if (Node.ELEMENT_NODE == child.getNodeType()) {
                        DocumentFragment fragment = doc.createDocumentFragment();
                        fragment.appendChild(child);
                        importedDocs.add(fragment);
                    } else {
                        child.getParentNode().removeChild(child);
                    }
                }
                importedDocumentsNode.getParentNode().removeChild(importedDocumentsNode);
            }
        }
        return importedDocs;
    }

    /**
     * Replace the service address in the wsdl:address element, not in the
     * ports.
     * 
     * @param doc
     * @param localAddress
     * @param newServiceName
     * @return
     * @throws TransformerException
     */
    public static final Document replaceAddressServiceIgnoreNS(final Document document,
            final String localAddress, final String newServiceName) throws TransformerException {
        Document doc = (Document) document.cloneNode(true);
        final List<Node> services = WSDLHelper.getChildNodes(doc.getFirstChild(), "service");
        for (final Node service : services) {
            final List<Node> ports = WSDLHelper.getChildNodes(service, "port");

            for (final Node port : ports) {
                final List<Node> addresses = WSDLHelper.getChildNodes(port, "address");
                for (final Node address : addresses) {
                    if (address instanceof Element) {
                        final Element a = (Element) address;
                        a.setAttribute("location", localAddress + "/" + newServiceName);
                    }
                }
            }
        }
        return doc;
    }

    /**
     * Get the children of the current with name ending with the given one
     * (ignore NS)
     * 
     * @param element
     * @param name
     * @return
     */
    private static final List<Node> getChildNodes(final Node element, final String name) {
        final List<Node> elements = new ArrayList<Node>();
        final NodeList children = element.getChildNodes();
        for (int i = 0; i < children.getLength(); i++) {
            final Node tmp = children.item(i);
            if (tmp.getNodeName().endsWith(name)) {
                elements.add(tmp);
            }
        }
        return elements;
    }

    /**
     * TODO : remove this method and use the EasyWSDL library instead.
     * 
     * Add the imported WSDl documents in the root document in the any part
     * 
     * @param doc
     * @param installroot
     * @param wsdlImportsDownload
     *            If <code>false</code>, the external imports declared in the
     *            service WSDL won't be downloaded, so they won't be replaced by
     *            their content.
     * @return
     * @throws PEtALSCDKException
     * @deprecated Use EasyWSDL API instead
     */
    @Deprecated
    @SuppressWarnings("unchecked")
    private static final org.jdom.Document addImportedDocuments(final org.jdom.Document doc,
            final String installroot, final boolean wsdlImportsDownload) throws PEtALSCDKException {
        if (doc != null) {
            final org.jdom.Element root = doc.getRootElement();
            final String wsdlRoot = root.getName();
            if (((wsdlRoot.equals("definitions")) && (root.getNamespaceURI().equals(WSDL11_NS)))
                    || ((wsdlRoot.equals("description")) && (root.getNamespaceURI()
                            .equals(WSDL20_NS)))) {

                final List<org.jdom.Element> imports = WSDLHelper.getAllImportedDocuments(doc,
                        installroot, wsdlImportsDownload);
                if ((imports != null) && (imports.size() > 0)) {
                    root.addNamespaceDeclaration(PETALS_WSDL_EXTENSIONS);
                    root.addNamespaceDeclaration(XSI);

                    if (root.getAttribute("schemaLocation", XSI) == null) {
                        root.setAttribute("schemaLocation",
                                "http://petals.ow2.org/wsdlExtensions wsdlExtensions.xsd", XSI);
                    } else {
                        String val = root.getAttributeValue("schemaLocation", XSI);
                        val = val + " http://petals.ow2.org/wsdlExtensions wsdlExtensions.xsd";
                        root.setAttribute("schemaLocation", val, XSI);
                    }

                    final org.jdom.Element importedDocumentsElmt = new org.jdom.Element(
                            "importedDocuments", PETALS_WSDL_EXTENSIONS);
                    for (final org.jdom.Element impt : imports) {
                        importedDocumentsElmt.addContent(impt);
                    }
                    root.getChildren().add(importedDocumentsElmt);
                }
            }
        }
        return doc;
    }

    /**
     * @param doc
     *            The master wsdl document to import
     * @param installroot
     *            The root master wsdl document path
     * @param wsdlImportsDownload
     *            If true, the imported document (wsdl or xsd)<br>
     *            mut be imported too. If false, they are ignored
     * @return a list of documents imported
     * @throws PEtALSCDKException
     * @deprecated Use EasyWSDL API instead
     */
    @Deprecated
    private static final List<org.jdom.Element> getAllImportedDocuments(
            final org.jdom.Document doc, final String installroot, final boolean wsdlImportsDownload)
            throws PEtALSCDKException {
        List<org.jdom.Element> res = null;
        String importedDocumentInstallroot = null;
        if (doc != null) {
            final org.jdom.Element root = doc.getRootElement();
            final String wsdlRoot = root.getName();
            if (((wsdlRoot.equals("definitions")) && (root.getNamespaceURI().equals(WSDL11_NS)))
                    || ((wsdlRoot.equals("description")) && (root.getNamespaceURI()
                            .equals(WSDL20_NS)))
                    || ((wsdlRoot.equals("schema")) && (root.getNamespaceURI().equals(SCHEMA_NS)))) {

                final List<org.jdom.Element> imports = getAllImports(root);

                // get in shemaLocation
                if (root.getAttributeValue("schemaLocation", XSI) != null) {
                    final String val = root.getAttributeValue("schemaLocation", XSI);
                    final StringTokenizer st = new StringTokenizer(val, " ");
                    String ns = null;
                    String location = null;
                    org.jdom.Element e = null;
                    while (st.hasMoreTokens()) {
                        ns = st.nextToken();
                        if (st.hasMoreTokens()) {
                            location = st.nextToken();
                            e = new org.jdom.Element("import", WSDL20);
                            e.setAttribute("schemaLocation", location);
                            e.setAttribute("namespace", ns);
                            imports.add(e);
                        }
                    }
                }

                if (imports.size() > 0) {
                    res = new ArrayList<org.jdom.Element>();
                    List<org.jdom.Element> inside = null;
                    String name = null;
                    String uri = null;
                    SAXBuilder builder = null;
                    org.jdom.Document importedDocument = null;
                    org.jdom.Element documentElmt = null;
                    URLConnection connection = null;
                    InputStream is = null;
                    try {
                        for (final org.jdom.Element impt : imports) {
                            name = impt.getAttributeValue("location");
                            if (name == null) {
                                name = impt.getAttributeValue("schemaLocation");
                            }
                            if (name != null) {
                                boolean isExternalURL = false;
                                try {
                                    final URL nameURL = new URL(name);
                                    final String protocol = nameURL.getProtocol();
                                    if ("http".equalsIgnoreCase(protocol)
                                            || "https".equalsIgnoreCase(protocol)) {
                                        isExternalURL = true;
                                    } else {
                                        isExternalURL = false;
                                    }
                                } catch (final MalformedURLException e1) {
                                    // The schemaLocation value is not an
                                    // external
                                    // URL because it is not wellformed
                                    isExternalURL = false;
                                }

                                if (!isExternalURL || (isExternalURL && wsdlImportsDownload)) {

                                    try {
                                        if (installroot != null) {
                                            uri = installroot + File.separator + name;
                                            final File f = new File(uri);
                                            importedDocumentInstallroot = f.getCanonicalFile()
                                                    .getParent();
                                            if (f.exists()) {
                                                connection = f.toURI().toURL().openConnection();
                                            } else {
                                                connection = new URI(name).toURL().openConnection();
                                            }
                                        } else {
                                            importedDocumentInstallroot = null;
                                            connection = new URI(name).toURL().openConnection();
                                        }
                                        is = connection.getInputStream();

                                        builder = new SAXBuilder();
                                        importedDocument = builder.build(is);

                                        if (importedDocument != null) {
                                            documentElmt = new org.jdom.Element("document",
                                                    PETALS_WSDL_EXTENSIONS);
                                            documentElmt.setAttribute("name", name);

                                            documentElmt
                                                    .addContent((org.jdom.Content) importedDocument
                                                            .getRootElement().clone());

                                            res.add(documentElmt);

                                            inside = WSDLHelper.getAllImportedDocuments(
                                                    importedDocument, importedDocumentInstallroot,
                                                    wsdlImportsDownload);

                                            if (inside != null) {
                                                for (final org.jdom.Element document : inside) {
                                                    res.add(document);
                                                }
                                            }
                                        }
                                    } catch (final MalformedURLException e) {
                                        // do nothing
                                        // the document is not included in
                                        // imported document
                                    } catch (final URISyntaxException e) {
                                        // do nothing
                                        // the document is not included in
                                        // imported document
                                    } catch (final UnknownHostException e) {
                                        // do nothing
                                        // the document is not included in
                                        // imported document
                                    } catch (final ConnectException e) {
                                        // do nothing
                                        // the document is not included in
                                        // imported document
                                        // TODO: This exception occurs when the
                                        // import is
                                        // an XSD available through an HTTP
                                        // connection and its
                                        // server is unacessible.
                                        // To prevent the server unacessibility
                                        // beacuse Internet
                                        // proxy, following parameters should be
                                        // added to the
                                        // CDK configuration: proxy host, proxy
                                        // port and non
                                        // proxied hosts (see feature#311110 on
                                        // the forge)
                                    }
                                }
                            }
                        }
                    } catch (final IOException e) {
                        throw new PEtALSCDKException(
                                "Can not get imported document for URL " + uri, e);
                    } catch (final JDOMException e) {
                        throw new PEtALSCDKException(
                                "Can not get imported document for URL " + uri, e);
                    }
                }
            }
        }
        return res;
    }

    /**
     * Get all the imports from the root element
     * 
     * @param root
     * @return
     * @deprecated Use EasyWSDL API instead
     */
    @SuppressWarnings("unchecked")
    @Deprecated
    private static final List<org.jdom.Element> getAllImports(final org.jdom.Element root) {
        final List<org.jdom.Element> imports = new ArrayList<org.jdom.Element>();

        // get in wsdl import or include
        for (final org.jdom.Element impt : (List<org.jdom.Element>) root.getChildren("import",
                WSDL11)) {
            imports.add((org.jdom.Element) impt);
        }
        for (final org.jdom.Element impt : (List<org.jdom.Element>) root.getChildren("import",
                WSDL20)) {
            imports.add((org.jdom.Element) impt);
        }
        for (final org.jdom.Element impt : (List<org.jdom.Element>) root.getChildren("include",
                WSDL20)) {
            imports.add((org.jdom.Element) impt);
        }

        // get in schema
        org.jdom.Element types = root.getChild("types", WSDL11);
        org.jdom.Element schema = null;
        if (types != null) {
            schema = types.getChild("schema", SCHEMA);
        } else {
            types = root.getChild("types", WSDL20);
            if (types != null) {
                schema = types.getChild("schema", SCHEMA);
            }
        }

        if (schema != null) {
            for (final org.jdom.Element impt : (List<org.jdom.Element>) schema.getChildren(
                    "import", SCHEMA)) {
                imports.add((org.jdom.Element) impt);
            }
            for (final org.jdom.Element impt : (List<org.jdom.Element>) schema.getChildren(
                    "include", SCHEMA)) {
                imports.add((org.jdom.Element) impt);
            }
        }
        return imports;
    }

}
